﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FeatureDetectors;
using System.Drawing;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.Structure;
using Khan08;
using DotNetMatrix;

namespace FeatureDetectors {

	public class KHANOptions : CDetectorOptions {
		public int wh = 5; // window size for thining edges
		public int N = 7; // minimal edge sequence
		public int s = 4;  // theta window size 
		public int z = 10;  // zero crossing window size
		public float zeroCroosRatio = 0.3f; // zero crossing threshold ratio
		public int zcS = 11; // zero crossing neibourgh window size
		public float pixelDistance = 5f;

		public float joinDistance = 4f;

		// displaying partial images
		public bool showEdges = false;
		public bool showGradientImage = false;
		public bool showEdgeSegments = false;
		
	}
	public class KHANFeature : Feature {

	}
	public class KHANPoint {
		public int X = 0;
		public int Y = 0;
		public double theta = 0; //radians
		public KHANPoint(int x, int y) {
			X = x;
			Y = y;
		}
		public Point ToPoint() {
			return new Point(X, Y);
		}
		public Feature ToFeature() {
			Feature res = new Feature();
			res.position = new Point(X, Y);
			res.location = new Point(X, Y);
			res.direction = theta;
			res.scale = 5;
			res.magnitude = 5;
			return res;
		}
	}
	public class CDetectorFunc : CDetector {
		public override string detectorName { get { return "KHAN08"; } }
		public override CDetectorOptions options { get; protected set; }
		public List<List<KHANPoint>> edges;
		public HashSet<int> activeEdges;
		public int[,] pointIndexer;
		public Random rand = new Random();
		public OptionsDialog optionsDialog;

		public CDetectorFunc() {
			options = new KHANOptions();
			optionsDialog = new Khan08.OptionsDialog((KHANOptions) options);
		}
		override public DialogResult ShowOptionsDialog() {

			if (optionsDialog != null) {
				optionsDialog.Show();
				optionsDialog.Activate();
			}
			return DialogResult.OK;
		}
		override public Feature[] GetFeatures(Image<Gray, byte> image) {
			Image<Gray, byte> originalImage = image.Clone();


			StructuringElementEx element = new StructuringElementEx(3, 3, 1, 1, Emgu.CV.CvEnum.CV_ELEMENT_SHAPE.CV_SHAPE_RECT);
			Image<Gray, byte> Iboc =  image.MorphologyEx(element, Emgu.CV.CvEnum.CV_MORPH_OP.CV_MOP_OPEN, 1);
			Iboc = Iboc.MorphologyEx(element, Emgu.CV.CvEnum.CV_MORPH_OP.CV_MOP_CLOSE, 1);

			Image<Gray, byte> Ibco =  image.MorphologyEx(element, Emgu.CV.CvEnum.CV_MORPH_OP.CV_MOP_CLOSE, 1);
			Ibco = Ibco.MorphologyEx(element, Emgu.CV.CvEnum.CV_MORPH_OP.CV_MOP_OPEN, 1);

			//Ib
			image = Iboc / 2 + Ibco / 2;
			

			//Ies
			image = image.MorphologyEx(element, Emgu.CV.CvEnum.CV_MORPH_OP.CV_MOP_GRADIENT, 1);



			if (((KHANOptions) options).showGradientImage) ShowImage(image);



			// THRESHOLDING
			float[,] g1f = { { -1, 0, 1 } };
			float[,] g2f = { { -1 }, { 0 }, { 1 } };
			ConvolutionKernelF g1 = new ConvolutionKernelF(g1f, new Point(1, 0));
			ConvolutionKernelF g2 = new ConvolutionKernelF(g2f, new Point(0, 1));
			Image<Gray, float> Ig1 = image.Convolution(g1);
			Image<Gray, float> Ig2 = image.Convolution(g2);
			//ShowImageInWindow(Ig1);
			//ShowImageInWindow(Ig2);
			Ig1 = Ig1.Max(Ig2);
			//ShowImageInWindow(Ig1);
			float sum1 = 0;
			float sum2 = 0;
			for (int x = 0; x<image.Width-1; x++)
				for (int y = 0; y < image.Height - 1; y++) {
					sum1 += Ig1.Data[y, x, 0] * image.Data[y, x, 0];
					sum2 += Ig1.Data[y, x, 0];
				}
			float gamma = sum2!=0 ? sum1 / sum2 : 0;
			image = image.ThresholdBinary(new Gray(gamma), new Gray(255));
			//ShowImageInWindow(image);


			// EDGE THINING
			Image<Gray, byte> Ive = image.Clone();
			for (int x = 0; x < Ive.Width; x++) {
				for (int y = 0; y < Ive.Height; y++) {
					if (image.Data[y, x, 0] != 0) {
						Ive.Data[y, x, 0] = 255;
						for (int k=1; k <= ((KHANOptions) options).wh/2; k++) {
							if (x - k >= 0) Ive.Data[y, x - k, 0] = 0;
							if (x + k < image.Width) Ive.Data[y, x + k, 0] = 0;
						}
					}
				}
			}

			Image<Gray, byte> Ihe = image.Clone();
			for (int y = 0; y < Ihe.Height; y++) {
				for (int x = 0; x < Ihe.Width; x++) {
					if (image.Data[y, x, 0] != 0) {
						Ihe.Data[y, x, 0] = 255;
						for (int k=1; k <= ((KHANOptions) options).wh/2; k++) {
							if (y - k >= 0) Ihe.Data[y - k, x, 0] = 0;
							if (y + k < image.Height) Ihe.Data[y + k, x, 0] = 0;
						}
					}
				}
			}

			
			image = Ive.Max(Ihe);

			if (((KHANOptions) options).showEdges) ShowImage(originalImage+image);

			BuildEdgeSegments(image);

			if (((KHANOptions) options).showEdgeSegments) {
				Image<Bgr, byte> lines = image.Convert<Bgr, byte>();
				int j = 0;
				foreach (int i in activeEdges) DrawEdge(edges[i], lines, j++);
				ShowImage(lines);
			}

			List<Feature> features = new List<Feature>();
			foreach (int i in activeEdges) {
				features.AddRange(GenerateFeatures(edges[i], originalImage));
			}

			// TEST
			/*
			ImageForm imgWnd = new ImageForm();
			Image<Bgr, byte> IMF = new Image<Bgr, byte>(Math.Max(5 * maxLength, image.Width), Math.Max(300, image.Height), new Bgr(255, 255, 255));
			imgWnd.imageBox1.Image = IMF;
			List<KHANPoint> edge = edges[maxEdge];
			int[] xs = new int[edge.Count];
			float[] xss = new float[edge.Count*4];
			float[] signal =  new float[edge.Count];
			Point[] pts = new Point[edge.Count];

			j = 0;
			for (int i = 0; i<xs.Length; i++) {
				xs[i] = i;
				xss[i * 4] = i;
				xss[i * 4 + 1] = i + 0.25f;
				xss[i * 4 + 2] = i + 0.50f;
				xss[i * 4 + 3] = i + 0.75f;
				signal[i] = (float) (edge[i].theta);
				pts[i] = edge[i].ToPoint();
			}
			float[] imf = EMD1dIMF1(signal, imgWnd);
			ShowPoints(xs, signal, imgWnd, new Bgr(255, 0, 0));
			ShowPoints(xs, imf, imgWnd, new Bgr(0, 0, 255));
			float[] spline = Spline(xs, imf, xss);
			ShowPoints(xss, spline , imgWnd, new Bgr(0, 0, 255));

			*/
			// END TEST


			//imgWnd.Show();


			
			return features.ToArray();
		}
		
		public void ShowImageInWindow(ImageForm imgWnd, Image<Gray, byte> image) {
			imgWnd.PicBox.Image = image;
			imgWnd.Show();
		}
		public void ClearScreen(ImageForm imgWnd) {
			Image<Bgr, byte> img = (Image<Bgr, byte>) imgWnd.PicBox.Image;
			imgWnd.PicBox.Image = new Image<Bgr, byte>(img.Width, img.Height, new Bgr(255, 255, 255));
		}
		public void ShowPoints(ImageForm imgWnd, Point[] pts, Bgr bgr) {
			Image<Bgr, byte> img = (Image<Bgr, byte>) imgWnd.PicBox.Image;
			img.DrawPolyline(pts, false, bgr, 1);
		}
		public void ShowPoints(int[] xs, float[] ys, ImageForm imgWnd, Bgr bgr) {
			Point[] pts = new Point[xs.Length];
			for (int i = 0; i < xs.Length; i++) {
				pts[i] = new Point(xs[i] * 5, 150 - (int) (ys[xs[i]]*50));
			}
			ShowPoints(imgWnd, pts, bgr);
		}
		public void ShowPoints(float[] xs, float[] ys, ImageForm imgWnd, Bgr bgr) {
			Point[] pts = new Point[xs.Length];
			for (int i = 0; i < xs.Length; i++) {
				pts[i] = new Point((int) (xs[i] * 5), 150 - (int) (ys[i] * 50));
			}
			ShowPoints(imgWnd, pts, bgr);
		}
		public Feature[] GenerateFeatures(List<KHANPoint> edge, Image<Gray, byte> image) {
			int N = edge.Count;
			float[] signal = new float[N];
			float[] xs = new float[N];
			float[] zero = new float[N];
			for (int i = 0; i < N; i++) {
				signal[i] = UpdateTheta(edge, i);
				xs[i] =i;
			}
			/*ImageForm imgWnd = new ImageForm();
			Image<Bgr, byte> freq  = new Image<Bgr, byte>(Math.Max(signal.Length*5, image.Width), Math.Max(300, image.Height), new Bgr(255, 255, 255));
			imgWnd.imageBox1.Image = freq;
			imgWnd.Show();
			*/

			float[] imf = EMD1dIMF1(signal, null);

			/*float[] diffS = new float[N]; 
			for (int i = 0; i < N; i++) diffS[i] = signal[i]-imf[i];
			ShowPoints(xs, zero, imgWnd, new Bgr(128, 128, 128));
			ShowPoints(xs, signal, imgWnd, new Bgr(255, 0, 0));
			ShowPoints(xs, imf, imgWnd, new Bgr(0, 0, 255));
			ShowPoints(xs, diffS, imgWnd, new Bgr(0, 255, 0));
			*/

			int z = ((KHANOptions) options).z;
			
			HashSet<int> candidates = new HashSet<int>();
			int IMFzeroCrossings = 0;
			for (int i = 0; i < N-1; i++) {
				if (imf[i]*imf[i+1] < 0) IMFzeroCrossings++;
			}
			float zeroCrossingThreshold = ((KHANOptions) options).zeroCroosRatio*IMFzeroCrossings;
			



			int[] 		zerocrossing	= new int[N];
			float[]		distances		= new float[N];


			for (int i = 0; i < N; i++) {
				for (int j = Math.Max(0,i - z/2); j <= Math.Min(N-2,i + z/2 - 1); j++) {
					// QUESTION: in article is imf[j-1]*imf[j+1
					if (imf[j]*imf[j+1] < 0) zerocrossing[i]++;
				}
				if (zerocrossing[i] > zeroCrossingThreshold) 
					candidates.Add(i);
			}
			int zcS = (int) ((KHANOptions)options).zcS;
			for (int i = 0; i < N; i++) {
				for (int j = Math.Max(0,i - zcS/2); j <= Math.Min(N-1,i + zcS/2); j++) {
					if (zerocrossing[j] != 0) distances[i] += Distance(edge[i], edge[j]);
				}
			}
			
			int[] candArr = candidates.ToArray();
			
			for (int i = 0; i < N; i++) {
				float minS = 99999;
				for (int j = Math.Max(0, i - zcS/2); j <= Math.Min(N-1, i + zcS/2); j++) {
					minS  = Math.Min(distances[j], minS);
				}
				for (int j = Math.Max(0, i - zcS/2); j <= Math.Min(N-1, i + zcS/2); j++) {
					if (zerocrossing[i]==zerocrossing[j] && distances[j]>minS) candidates.Remove(j);
				}	
			}

			candArr = candidates.ToArray();
			for (int i = 0; i < candArr.Length; i++) {
				if (!IsLocalExtremum(edge[candArr[i]].ToPoint(), image)) candidates.Remove(candArr[i]);
			}

			candArr = candidates.ToArray();
			float dist = ((KHANOptions)options).pixelDistance;
			for (int i = 1; i < candArr.Length; i++) {
				if (Distance(edge[candArr[i]], edge[candArr[i-1]]) < dist) {
					candidates.Remove(candArr[i]);
					candArr[i] = candArr[i-1];
				}
			}
			
			Feature[] res = new Feature[candidates.Count];
			int k = 0;
			foreach (int i in candidates) {
				res[k++] = edge[i].ToFeature();
			}

			return res;
		}

		public float Distance(KHANPoint p1, KHANPoint p2) {
			return (float) Math.Sqrt(Math.Pow(p1.X-p2.X, 2) + Math.Pow(p1.Y-p2.Y, 2));
		}
		public bool IsLocalExtremum(Point p, Image<Gray, byte> image) {
			byte min = 255, max = 0;
			for (int x = Math.Max(0, p.X - 1); x <= Math.Min(image.Width - 1, p.X + 1); x++)
				for (int y = Math.Max(0, p.Y - 1); y <= Math.Min(image.Height-1, p.Y + 1); y++) {
					min = Math.Min(min, image.Data[y, x, 0]);
					max = Math.Max(max, image.Data[y, x, 0]);
					if (image.Data[p.Y, p.X, 0] > min && image.Data[p.Y, p.X, 0] < max) return false;
				}
			return true;
		}
		public void BuildEdgeSegments(Image<Gray, byte> image) {
			edges = new List<List<KHANPoint>>();
			pointIndexer = new int[image.Height, image.Width];
			activeEdges = new HashSet<int>();

			for (int y = 0; y < image.Height; y++) {
				for (int x = 0; x < image.Width; x++) {
					if (pointIndexer[y, x] == 0 && image.Data[y, x, 0] != 0) {


						if (x>0 && pointIndexer[y, x - 1] != 0) {
							pointIndexer[y, x] = pointIndexer[y, x - 1];
							AddPointToSegment(pointIndexer[y, x - 1], new KHANPoint(x, y));
						}
						if (y > 0) {

							if (x > 0 && pointIndexer[y - 1, x - 1] != 0) {
								if (pointIndexer[y, x] != 0) MergeEdges(pointIndexer[y, x], pointIndexer[y - 1, x - 1]);
								else {
									pointIndexer[y, x] = pointIndexer[y - 1, x - 1];
									AddPointToSegment(pointIndexer[y - 1, x - 1], new KHANPoint(x, y));
								}
							}
							if (x == image.Width - 1 || image.Data[y, x + 1, 0] == 0) {
								if (pointIndexer[y - 1, x] != 0) {
									if (pointIndexer[y, x] != 0) MergeEdges(pointIndexer[y, x], pointIndexer[y - 1, x]);
									else {
										pointIndexer[y, x] = pointIndexer[y - 1, x];
										AddPointToSegment(pointIndexer[y - 1, x], new KHANPoint(x, y));
									}
								}
								if (x < image.Width - 1 && pointIndexer[y - 1, x + 1] != 0) {
									if (pointIndexer[y, x] != 0) MergeEdges(pointIndexer[y, x], pointIndexer[y - 1, x + 1]);
									else {
										pointIndexer[y, x] = pointIndexer[y - 1, x + 1];
										AddPointToSegment(pointIndexer[y - 1, x + 1], new KHANPoint(x, y));
									}
								}
							}
						}
						if (pointIndexer[y, x] == 0) {
							CreateLineSegment(new KHANPoint(x, y));
						}
					}
				}
			}


			int [] temp = activeEdges.ToArray();
			foreach (int i in temp) {
				if (edges[i].Count < ((KHANOptions) options).N) RemoveEdge(i, image);
			}

			JoinLineSegments();
		}
		public void JoinLineSegments() {
			int[] active = activeEdges.ToArray();
			float dist = ((KHANOptions)options).joinDistance;
			for (int i=0; i<active.Length-1; i++) {
				for (int j=i+1; j<active.Length; j++) {
					if (i==j || active[i]<0 || active[j]<0) continue;
					int e1 = active[i];
					int e2 = active[j];
					int cnt1 = edges[e1].Count;
					int cnt2 = edges[e2].Count;
					if (Distance(edges[e1][cnt1-1], edges[e2][0]) <= dist || Distance(edges[e1][cnt1-1], edges[e2][cnt2-1]) <= dist) {
						MergeEdges(e1, e2);
						active[j] = -1;
					}
					else if (Distance(edges[e2][cnt2-1], edges[e1][cnt1-1]) <= dist || Distance(edges[e2][cnt2-1], edges[e1][0]) <= dist) {
						MergeEdges(e2, e1);
						active[i] = -1;
					}
				}
			}
		}
		public void CreateLineSegment(KHANPoint point) {
			pointIndexer[point.Y, point.X] = edges.Count;
			edges.Add(new List<KHANPoint>());
			activeEdges.Add(pointIndexer[point.Y, point.X]);
			edges[pointIndexer[point.Y, point.X]].Add(point);
		}
		public void AddPointToSegment(int edgeIndex, KHANPoint point) {
			KHANPoint p = edges[edgeIndex].Last<KHANPoint>();
			KHANPoint s = edges[edgeIndex][0];
			if (Math.Abs(point.X - s.X) <= 1 && Math.Abs(point.Y - s.Y) <= 1) {
				edges[edgeIndex].Reverse();
			}
			else if (Math.Abs(point.X - p.X) > 1 || Math.Abs(point.Y - p.Y) > 1) {
				CreateLineSegment(point);
				return;
			}
			edges[edgeIndex].Add(point);
			pointIndexer[point.Y, point.X] = edgeIndex;
		}

		// e1 actual scaning edge
		public void MergeEdges(int e1, int e2) {
			if (e1 == e2) return ;

			KHANPoint Emax = edges[e1].Last<KHANPoint>();
			KHANPoint Emin = edges[e2].Last<KHANPoint>();
			KHANPoint Smin = edges[e2][0];

			if (Math.Abs(Emax.X -Emin.X)<=1 && Math.Abs(Emax.Y -Emin.Y) <= 1) {
				edges[e2].Reverse();
			}
			edges[e1].AddRange(edges[e2]);

			foreach (KHANPoint ff in edges[e2]) pointIndexer[ff.Y, ff.X] = e1;

			edges[e2] = null;
			activeEdges.Remove(e2);
		}
		public void RemoveEdge(int edgeIndex, Image<Gray, byte> image) {

			activeEdges.Remove(edgeIndex);
			if (edges[edgeIndex] == null) return;
			foreach (KHANPoint ff in edges[edgeIndex]) {
				pointIndexer[ff.Y, ff.X] = 0;
				image.Data[ff.Y, ff.X, 0] = 0;
			}
			edges[edgeIndex] = null;
		}
		public void DrawEdge(List<KHANPoint> edge, Image<Bgr, byte> img, int index) {
			Bgr bgr = new Bgr(rand.NextDouble() * 255, rand.NextDouble() * 255, rand.NextDouble() * 255);
			//	img.DrawPolyline(edge.ToArray(), false, new Bgr(0, 255, 0), 1);
			for (int i = 1; i < edge.Count-1; i++) {
				Bgr color = new Bgr(128 + bgr.Blue * i / edge.Count / 2, 128 + bgr.Green * i / edge.Count / 2, 128 +  bgr.Red * i / edge.Count / 2);
				img.Draw(new LineSegment2D(edge[i].ToPoint(), edge[i].ToPoint()), color, 1);
			}
			img.Draw(new LineSegment2D(edge[0].ToPoint(), edge[0].ToPoint()), new Bgr(255, 255, 255), 1);
			img.Draw(new LineSegment2D(edge[edge.Count - 1].ToPoint(), edge[edge.Count - 1].ToPoint()), new Bgr(0, 255, 255), 1);

		}
		public GeneralMatrix FunctionM(List<KHANPoint> edge, int index) {
			int s = ((KHANOptions) options).s;
			s = Math.Min(s, Math.Min(edge.Count - index - 1, index));
			float konst = (float) 1/(2*s+1);
			float sum1 = 0, sum2 = 0, sum3 = 0, Xi = 0, Yi = 0;
			for (int i = index - s; i <= index+s; i++) {
				if (i < 0 || i >= edge.Count) continue;
				sum1 += edge[i].X* edge[i].X;
				sum2 += edge[i].Y* edge[i].Y;
				sum3 += edge[i].X* edge[i].Y;

				Xi += edge[i].X;
				Yi += edge[i].Y;
			}
			sum1*= konst;
			sum2*= konst;
			sum3*= konst;
			Xi*= konst;
			Yi*= konst;

			GeneralMatrix M = new GeneralMatrix(2, 2);
			M[0, 0] = sum1 - Xi*Xi;
			M[0, 1] = sum3 - Xi*Yi;
			M[1, 0] = M[0, 1];
			M[1, 1] = sum2 - Yi*Yi;
			return M;
		}

		public float UpdateTheta(List<KHANPoint> edge, int index) {
			GeneralMatrix M = FunctionM(edge, index);

			EigenvalueDecomposition xx = M.Eigen();
			double eigen1 = 0;
			if (xx.RealEigenvalues.Length > 0) eigen1 = xx.RealEigenvalues[0];
			double theta = Math.PI / 2;
			if (M[0, 1] != 0) theta = Math.Abs(Math.Atan((eigen1 - M[0, 0]) / M[0, 1]));
			return (float) (edge[index].theta = theta);
		}

		public float[] EMD1dIMF1(float[] signal, ImageForm imgWnd) {

			/*
			 * EMD:  Emprical mode decomposition
			 * 
			 *
			 * See:  Huang et al, Royal Society Proceedings on Math, Physical, 
			 *       and Engineering Sciences, vol. 454, no. 1971, pp. 903-995, 
			 *       8 March 1998
			 *
			 * Author: Ivan Magrin-Chagnolleau  <ivan@ieee.org>
			 * Transcribed to C#: Igor Kolenic <fuuzix@gmail.com>
			 */


			float[] c = signal.ToArray();
			int N = signal.Length;
			float maxAmplitude = 0;
			foreach (float f in signal) maxAmplitude = Math.Max(maxAmplitude, Math.Abs(f));

			float[] h = c.ToArray();

			int iteration = 0, maxIterations = 100;
			float SD = 1;
			while ((SD > 0.3)  && iteration++ < maxIterations) { // % while the standard deviation is higher than 0.3 (typical value)
				// find local max/min points
				// approximate derivative


				/*// DRAWING
				float[] xs = new float[h.Length];
				for (int i = 0; i< h.Length; i++) xs[i] = i;
				ClearScreen(imgWnd);
				ShowPoints(xs, h, imgWnd, new Bgr(0, 0, 0));
				*/

				float[][] extremes = GetPeaks(h);
				if (extremes == null) break;


				// spline interpolate to get max and min envelopes; form imf
				float[] xx = new float[N];

				for (int i = 0; i< N; i++) xx[i] = i;
				float[] maxenv = Spline(extremes[0], extremes[1], xx);
				float[] minenv = Spline(extremes[2], extremes[3], xx);


				/*/ DRAWING
				ShowPoints(xx, minenv, imgWnd, new Bgr(0, 0, 255));
				ShowPoints(xx, maxenv, imgWnd, new Bgr(0, 255, 255));
				//MessageBox.Show("EXTREMES", "XX", MessageBoxButtons.OK, MessageBoxIcon.Asterisk);
				 */
				
				// m - mean of max and min enveloppes
				float[] m = new float[N];
				float[] hn = new float[N];
				SD = 0;
				float sum1 = 0, sum2 = 0;
				for (int i = 0; i< N; i++) {
					m[i] = (maxenv[i] + minenv[i])/2;
					hn[i] = h[i] - m[i]; //substract mean to h
					// calculate standard deviation
					SD += (m[i]*m[i]) / (h[i]*h[i]+ 0.0000001f);
				}

				/*/ DRAWING
				ShowPoints(xx, m, imgWnd, new Bgr(0, 255, 0));
				ShowPoints(xx, hn, imgWnd, new Bgr(200, 200, 200));
				imgWnd.Hide();
				imgWnd.ShowDialog();
				*/

				//SD = sum1 / (sum2 + 0.0000001f);
				h = hn;
			}
			return h;
		}
		public float[][] GetPeaks(float[] signal) {
			int N = signal.Length;

			short[] d = new short[N+1];
			for (int i = 0; i< N-1; i++)  d[i] = (short) Math.Sign(signal[i+1] - signal[i]);

			List<int> maxmin = new List<int>(); //to store the optima (min and max without distinction so far)
			maxmin.Add(0);
			for (int i = 1; i < N - 1; i++) {
				if (d[i] == 0) {  	 // we are on a zero
					if (d[i-1] != d[i+1]) {  // it is a maximum
						maxmin.Add(i);
					}
				}
				else if (d[i] != d[i+1]) {   // we are straddling a zero so
					maxmin.Add(i+1);        // define zero as at i+1 (not i)
				}
			}
			if (maxmin.Last() != N-1) maxmin.Add(N-1);


			if (maxmin.Count <= 3) return null;
			// divide maxmin into maxes and mins
			bool firstmax = maxmin[0]>maxmin[1];
			bool odd = maxmin.Count % 2 != 0;
			int maxcnt = maxmin.Count / 2 + 1 + ((!firstmax && odd) ? 1 : 0);
			int mincnt = maxmin.Count / 2 + 1 + ((firstmax && !odd) ? 1 : 0);
			float[][] res  = new float[4][];
			res[0] = new float[maxcnt];
			res[1] = new float[maxcnt];
			res[2] = new float[mincnt];
			res[3] = new float[mincnt];


			if (firstmax) {
				for (int i = 0; i<maxmin.Count; i += 2) {
					res[0][i/2] = maxmin[i];
					res[1][i/2] = signal[maxmin[i]];
				}
				for (int i = 1; i<maxmin.Count; i += 2) {
					res[2][i/2 +1] = maxmin[i];
					res[3][i/2 +1] = signal[maxmin[i]];
				}
				res[2][0] = 0;
				res[3][0] = res[3][1];
				if (odd) {
					res[2][mincnt-1] = N - 1;
					res[3][mincnt-1] = res[3][mincnt-2];
				}
				else {
					res[0][maxcnt-1] = N - 1; ;
					res[1][maxcnt-1] = res[1][maxcnt-2];
				}
			}
			else {

				for (int i = 1; i<maxmin.Count; i += 2) {
					res[0][i/2+1] = maxmin[i];
					res[1][i/2+1] = signal[maxmin[i]];
				}
				for (int i = 0; i<maxmin.Count; i += 2) {
					res[2][i/2] = maxmin[i];
					res[3][i/2] = signal[maxmin[i]];
				}
				res[0][0] = 0;
				res[1][0] = res[1][1];

				if (!odd) {
					res[2][mincnt-1] = N - 1;
					res[3][mincnt-1] = res[3][mincnt-2];
				}
				else {
					res[0][maxcnt-1] = N - 1;
					res[1][maxcnt-1] = res[1][maxcnt-2];
				}

			}
			return res;
		}
		public float[] Spline(int[] x, float[] y, float[] newx) {
			float[] yi = new float[x.Length];
			float[] xi = new float[x.Length];
			for (int i=0; i< x.Length; i++) {
				yi[i] = y[x[i]];
				xi[i] = x[i];
			}
			return Spline(xi, yi, newx);
		}
		public float[] Spline(float[] xi, float[] yi, float[] newx) {

			float[,] S = GetSplainFunctions(xi, yi);

			float[] res = new float[newx.Length];
			int j = 0;
			for (int i = 0; i < newx.Length; i++) {
				while (j<xi.Length-1 && newx[i]>=xi[j+1]) j++;

				float dif = (newx[i] - S[j, 0]);
				res[i] = S[j, 1] + S[j, 2]*dif + S[j, 3]*dif*dif + S[j, 4]*dif*dif*dif;

			}
			return res;
		}


		public float[,] GetSplainFunctions(float[] xi, float[] fi, bool nak = false) {
			int n = xi.Length;
			int N = n - 1; // number of intervals
			float[] h = new float[N];
			for (int i = 0; i < N; i++) h[i] = xi[i + 1] - xi[i];
			float[,] res = new float[n, 5];

			float[] dd = new float[N-1];
			float[] ri = new float[N-1];

			for (int i = 0; i < N-1; i++) {
				dd[i] = 2.0f * (h[i] + h[i + 1]);
				ri[i] = 6.0f * ((fi[i + 2] - fi[i + 1]) / h[i + 1] - (fi[i + 1] - fi[i]) / h[i]);
			}


			float[] du = new float[N - 2];
			for (int i = 0; i < N - 2; i++) du[i] = h[i + 1];
			float[] dl = du.ToArray();

			float[] temp = TridiagonalSolve(dl, dd, du, ri);


			float[] S = new float[n];
			for (int i = 1; i < N; i++) S[i] = temp[i - 1];

			for (int i = 0; i < n; i++) {
				res[i, 0] = xi[i];
				res[i, 1] = fi[i];
			}
			for (int i = 0; i < N; i++) {
				res[i, 2] = (fi[i + 1] - fi[i]) / h[i] - (2 * h[i] * S[i] + h[i] * S[i + 1]) / 6.0f;
				res[i, 3] = S[i] / 2.0f;

				res[i, 4] = (S[i+1] - S[i])/ (6.0f*h[i]);
			}


			return res;
		}

		public float[] TridiagonalSolve(float[] a, float[] b, float[] c, float[] d) {
			/* method source from http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm */
			/* Modify the coefficients. */
			int n = d.Length;
			float[] res;

			if (n<2) {
				res = new float[1];
				res[0] =  d[0] / b[0];
				return res;
			}
			c[0] /= b[0];	/* Division by zero risk. */
			d[0] /= b[0];	/* Division by zero would imply a singular matrix. */
			for (int i = 1; i < n-1; i++) {
				float id = 1 / (b[i] - c[i-1] * a[i-1]);  /* Division by zero risk. */
				if (i<n-1) c[i] *= id;	                         /* Last value calculated is redundant. */
				d[i] = (d[i] - d[i-1] * a[i-1]) * id;
			}
			d[n-1] = (d[n-1] - d[n-2] * a[n-2]) * c[n-2];

			res = new float[n];
			/* Now back substitute. */
			res[n - 1] = d[n - 1];
			for (int i = n - 2; i >= 0; i--)
				res[i] = d[i] - c[i] * res[i + 1];

			return res;

		}

	}
}
